@@ -5,3 +5,6 @@ mod hash_table;
 
 pub mod hash_map;
 pub use hash_map::HashMap;
+
+pub mod tree_index;
+pub use tree_index::TreeIndex;
