@@ -1,12 +1,18 @@
+use super::wait_queue::WaitQueue;
+
 use crate::ebr::{Arc, AtomicArc, Barrier, Ptr, Tag};
 
 use std::borrow::Borrow;
 use std::mem::MaybeUninit;
 use std::ptr;
 use std::sync::atomic::fence;
+use std::sync::atomic::AtomicU32;
 use std::sync::atomic::Ordering::{Acquire, Relaxed, Release};
-use std::sync::atomic::{AtomicPtr, AtomicU32};
-use std::sync::{Condvar, Mutex};
+
+/// The fixed size of [`DataArray`].
+///
+/// The size cannot exceed `32`.
+pub const ARRAY_SIZE: usize = 32;
 
 /// State bits.
 const KILLED: u32 = 1_u32 << 31;
@@ -17,9 +23,9 @@ const LOCK_MASK: u32 = LOCK | SLOCK_MAX;
 
 /// [`Cell`] is a small fixed-size hash table that resolves hash conflicts using a linked list
 /// of entry arrays.
-pub struct Cell<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> {
-    /// `wait_queue` additionally stores the state of the [`Cell`]: locked or killed.
-    wait_queue: AtomicPtr<WaitQueueEntry>,
+pub struct Cell<K: 'static + Eq, V: 'static, const LOCK_FREE: bool> {
+    /// An array of key-value pairs and their metadata.
+    data_array: AtomicArc<DataArray<K, V>>,
 
     /// The state of the [`Cell`].
     state: AtomicU32,
@@ -27,26 +33,22 @@ pub struct Cell<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE:
     /// The number of valid entries in the [`Cell`].
     num_entries: u32,
 
-    /// DataArray stores key-value pairs with their metadata.
-    data: AtomicArc<DataArray<K, V, SIZE>>,
+    /// The wait queue of the [`Cell`].
+    wait_queue: WaitQueue,
 }
 
-impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> Default
-    for Cell<K, V, SIZE, LOCK_FREE>
-{
+impl<K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Default for Cell<K, V, LOCK_FREE> {
     fn default() -> Self {
-        Cell::<K, V, SIZE, LOCK_FREE> {
-            wait_queue: AtomicPtr::default(),
+        Cell::<K, V, LOCK_FREE> {
+            data_array: AtomicArc::null(),
             state: AtomicU32::new(0),
             num_entries: 0,
-            data: AtomicArc::null(),
+            wait_queue: WaitQueue::default(),
         }
     }
 }
 
-impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
-    Cell<K, V, SIZE, LOCK_FREE>
-{
+impl<K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Cell<K, V, LOCK_FREE> {
     /// Returns true if the [`Cell`] has been killed.
     pub fn killed(&self) -> bool {
         (self.state.load(Relaxed) & KILLED) == KILLED
@@ -58,8 +60,8 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     }
 
     /// Iterates the contents of the [`Cell`].
-    pub fn iter<'b>(&'b self, guard: &'b Barrier) -> EntryIterator<'b, K, V, SIZE, LOCK_FREE> {
-        EntryIterator::new(self, guard)
+    pub fn iter<'b>(&'b self, barrier: &'b Barrier) -> EntryIterator<'b, K, V, LOCK_FREE> {
+        EntryIterator::new(self, barrier)
     }
 
     /// Searches for an entry associated with the given key.
@@ -78,8 +80,8 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
             return None;
         }
 
-        let mut data_array_ptr = self.data.load(Relaxed, barrier);
-        let preferred_index = partial_hash as usize % SIZE;
+        let mut data_array_ptr = self.data_array.load(Relaxed, barrier);
+        let preferred_index = partial_hash as usize % ARRAY_SIZE;
         while let Some(data_array_ref) = data_array_ptr.as_ref() {
             if let Some((_, entry_ref)) =
                 Self::search_array(data_array_ref, key_ref, preferred_index, partial_hash)
@@ -91,14 +93,14 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
         None
     }
 
-    /// Gets a [`EntryIterator`] pointing to an entry associated with the given key.
+    /// Gets an [`EntryIterator`] pointing to an entry associated with the given key.
     #[inline]
     pub fn get<'b, Q>(
         &'b self,
         key_ref: &Q,
         partial_hash: u8,
         barrier: &'b Barrier,
-    ) -> Option<EntryIterator<'b, K, V, SIZE, LOCK_FREE>>
+    ) -> Option<EntryIterator<'b, K, V, LOCK_FREE>>
     where
         K: Borrow<Q>,
         Q: Eq + ?Sized,
@@ -107,11 +109,9 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
             return None;
         }
 
-        // In order to read the linked list correctly, a load-acquire is required.
-        let read_order = if LOCK_FREE { Acquire } else { Relaxed };
-        let mut current_array_ptr = self.data.load(read_order, barrier);
+        let mut current_array_ptr = self.data_array.load(Relaxed, barrier);
         let mut prev_array_ptr = Ptr::null();
-        let preferred_index = partial_hash as usize % SIZE;
+        let preferred_index = partial_hash as usize % ARRAY_SIZE;
         while let Some(data_array_ref) = current_array_ptr.as_ref() {
             if let Some((index, _)) =
                 Self::search_array(data_array_ref, key_ref, preferred_index, partial_hash)
@@ -125,7 +125,7 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
                 });
             }
             prev_array_ptr = current_array_ptr;
-            current_array_ptr = data_array_ref.link.load(read_order, barrier);
+            current_array_ptr = data_array_ref.link.load(Relaxed, barrier);
         }
         None
     }
@@ -133,7 +133,7 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     /// Searches the given [`DataArray`] for an entry matching the key.
     #[inline]
     fn search_array<'b, Q>(
-        data_array_ref: &'b DataArray<K, V, SIZE>,
+        data_array_ref: &'b DataArray<K, V>,
         key_ref: &Q,
         preferred_index: usize,
         partial_hash: u8,
@@ -165,7 +165,7 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
 
         // Looks into other slots.
         let mut current_index = occupied.trailing_zeros();
-        while (current_index as usize) < SIZE {
+        while (current_index as usize) < ARRAY_SIZE {
             let entry_ptr = data_array_ref.data[current_index as usize].as_ptr();
             let entry_ref = unsafe { &(*entry_ptr) };
             if entry_ref.0.borrow() == key_ref {
@@ -182,14 +182,14 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     ///
     /// If the given slot is valid, it returns the given slot.
     fn next_entry<Q>(
-        data_array_ref: &DataArray<K, V, SIZE>,
+        data_array_ref: &DataArray<K, V>,
         current_index: usize,
     ) -> Option<(usize, &(K, V), u8)>
     where
         K: Borrow<Q>,
         Q: Eq + ?Sized,
     {
-        if current_index >= SIZE {
+        if current_index >= ARRAY_SIZE {
             return None;
         }
 
@@ -205,7 +205,7 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
         }
 
         let next_index = occupied.trailing_zeros() as usize;
-        if next_index < SIZE {
+        if next_index < ARRAY_SIZE {
             let entry_ptr = data_array_ref.data[next_index].as_ptr();
             return Some((
                 next_index,
@@ -216,74 +216,29 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
 
         None
     }
-
-    /// Waits for the owner thread to release the [`Cell`].
-    fn wait<T, F: FnOnce() -> Option<T>>(&self, f: F) -> Option<T> {
-        // Inserts the condvar into the wait queue.
-        let mut current = self.wait_queue.load(Relaxed);
-        let mut entry = WaitQueueEntry::new(current);
-
-        while let Err(actual) = self.wait_queue.compare_exchange(
-            current,
-            &mut entry as *mut WaitQueueEntry,
-            Release,
-            Relaxed,
-        ) {
-            current = actual;
-            entry.next_ptr = current;
-        }
-
-        // Marks that there is a waiting thread.
-        self.state.fetch_or(WAITING, Release);
-
-        // Tries to lock again once the entry is inserted into the wait queue.
-        let locked = f();
-        if locked.is_some() {
-            self.signal();
-        }
-
-        // Locking failed.
-        entry.wait();
-        locked
-    }
-
-    /// Signals the threads in the wait queue.
-    fn signal(&self) {
-        let mut current = self.wait_queue.swap(ptr::null_mut(), Acquire);
-        while let Some(entry_ref) = unsafe { current.as_ref() } {
-            let next_ptr = entry_ref.next_ptr;
-            entry_ref.signal();
-            current = next_ptr;
-        }
-    }
 }
 
-impl<K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> Drop
-    for Cell<K, V, SIZE, LOCK_FREE>
-{
+impl<K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Drop for Cell<K, V, LOCK_FREE> {
     fn drop(&mut self) {
         // The Cell must have been killed.
         debug_assert!(self.killed());
     }
 }
 
-pub struct EntryIterator<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
-{
-    cell_ref: Option<&'b Cell<K, V, SIZE, LOCK_FREE>>,
-    current_array_ptr: Ptr<'b, DataArray<K, V, SIZE>>,
-    prev_array_ptr: Ptr<'b, DataArray<K, V, SIZE>>,
+pub struct EntryIterator<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> {
+    cell_ref: Option<&'b Cell<K, V, LOCK_FREE>>,
+    current_array_ptr: Ptr<'b, DataArray<K, V>>,
+    prev_array_ptr: Ptr<'b, DataArray<K, V>>,
     current_index: usize,
     barrier_ref: &'b Barrier,
 }
 
-impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
-    EntryIterator<'b, K, V, SIZE, LOCK_FREE>
-{
+impl<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> EntryIterator<'b, K, V, LOCK_FREE> {
     /// Creates a new [`EntryIterator`].
     pub fn new(
-        cell: &'b Cell<K, V, SIZE, LOCK_FREE>,
+        cell: &'b Cell<K, V, LOCK_FREE>,
         barrier: &'b Barrier,
-    ) -> EntryIterator<'b, K, V, SIZE, LOCK_FREE> {
+    ) -> EntryIterator<'b, K, V, LOCK_FREE> {
         EntryIterator {
             cell_ref: Some(cell),
             current_array_ptr: Ptr::null(),
@@ -307,7 +262,7 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     ///
     /// It should only be invoked when the caller is holding a [`Locker`] on the [`Cell`].
     #[inline]
-    fn unlink_data_array(&mut self, data_array_ref: &DataArray<K, V, SIZE>) {
+    fn unlink_data_array(&mut self, data_array_ref: &DataArray<K, V>) {
         let next_data_array = if LOCK_FREE {
             data_array_ref.link.get_arc(Relaxed, self.barrier_ref)
         } else {
@@ -321,8 +276,13 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
                 .link
                 .swap((next_data_array, Tag::None), Relaxed)
         } else if let Some(cell_ref) = self.cell_ref.as_ref() {
-            debug_assert!(!cell_ref.data.load(Relaxed, self.barrier_ref).is_null());
-            cell_ref.data.swap((next_data_array, Tag::None), Relaxed)
+            debug_assert!(!cell_ref
+                .data_array
+                .load(Relaxed, self.barrier_ref)
+                .is_null());
+            cell_ref
+                .data_array
+                .swap((next_data_array, Tag::None), Relaxed)
         } else {
             None
         };
@@ -337,16 +297,15 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     }
 }
 
-impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> Iterator
-    for EntryIterator<'b, K, V, SIZE, LOCK_FREE>
+impl<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Iterator
+    for EntryIterator<'b, K, V, LOCK_FREE>
 {
     type Item = (&'b (K, V), u8);
     fn next(&mut self) -> Option<Self::Item> {
         if let Some(&cell_ref) = self.cell_ref.as_ref() {
-            let read_order = if LOCK_FREE { Acquire } else { Relaxed };
             if self.current_array_ptr.is_null() {
                 // Starts scanning from the beginning.
-                self.current_array_ptr = cell_ref.data.load(read_order, self.barrier_ref);
+                self.current_array_ptr = cell_ref.data_array.load(Relaxed, self.barrier_ref);
             }
             while let Some(data_array_ref) = self.current_array_ptr.as_ref() {
                 // Searches for the next valid entry.
@@ -356,7 +315,7 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
                     self.current_index + 1
                 };
                 if let Some((index, entry_ref, hash)) =
-                    Cell::<K, V, SIZE, LOCK_FREE>::next_entry(data_array_ref, current_index)
+                    Cell::<K, V, LOCK_FREE>::next_entry(data_array_ref, current_index)
                 {
                     self.current_index = index;
                     return Some((entry_ref, hash));
@@ -364,7 +323,7 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
 
                 // Proceeds to the next DataArray.
                 self.prev_array_ptr = self.current_array_ptr;
-                self.current_array_ptr = data_array_ref.link.load(read_order, self.barrier_ref);
+                self.current_array_ptr = data_array_ref.link.load(Relaxed, self.barrier_ref);
                 self.current_index = usize::MAX;
             }
             // Fuses itself.
@@ -374,29 +333,27 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     }
 }
 
-pub struct Locker<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> {
-    cell_ref: &'b Cell<K, V, SIZE, LOCK_FREE>,
-    killed: bool,
+pub struct Locker<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> {
+    cell_ref: &'b Cell<K, V, LOCK_FREE>,
 }
 
-impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SIZE, LOCK_FREE> {
+impl<'b, K: Eq, V, const LOCK_FREE: bool> Locker<'b, K, V, LOCK_FREE> {
     /// Locks the given [`Cell`].
     #[inline]
     pub fn lock(
-        cell: &'b Cell<K, V, SIZE, LOCK_FREE>,
+        cell: &'b Cell<K, V, LOCK_FREE>,
         barrier: &'b Barrier,
-    ) -> Option<Locker<'b, K, V, SIZE, LOCK_FREE>> {
+    ) -> Option<Locker<'b, K, V, LOCK_FREE>> {
         loop {
-            if let Some(locker) = Self::try_lock(cell, barrier)
-                .map_or_else(|| cell.wait(|| Self::try_lock(cell, barrier)), Some)
-            {
-                if locker.killed {
-                    return None;
-                }
-                return Some(locker);
+            if let Ok(locker) = Self::try_lock(cell, barrier) {
+                return locker;
             }
-            if cell.killed() {
-                return None;
+            if let Ok(locker) = cell.wait_queue.wait(|| {
+                // Marks that there is a waiting thread.
+                cell.state.fetch_or(WAITING, Release);
+                Self::try_lock(cell, barrier)
+            }) {
+                return locker;
             }
         }
     }
@@ -404,50 +361,46 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
     /// Tries to lock the [`Cell`].
     #[inline]
     fn try_lock(
-        cell: &'b Cell<K, V, SIZE, LOCK_FREE>,
+        cell: &'b Cell<K, V, LOCK_FREE>,
         _barrier: &'b Barrier,
-    ) -> Option<Locker<'b, K, V, SIZE, LOCK_FREE>> {
+    ) -> Result<Option<Locker<'b, K, V, LOCK_FREE>>, ()> {
         let current = cell.state.load(Relaxed) & (!LOCK_MASK);
+        if (current & KILLED) == KILLED {
+            return Ok(None);
+        }
         if cell
             .state
             .compare_exchange(current, current | LOCK, Acquire, Relaxed)
             .is_ok()
         {
-            Some(Locker {
-                cell_ref: cell,
-                killed: (current & KILLED) == KILLED,
-            })
+            Ok(Some(Locker { cell_ref: cell }))
         } else {
-            None
+            Err(())
         }
     }
 
     /// Returns a reference to the [`Cell`].
     #[inline]
-    pub fn cell_ref(&self) -> &'b Cell<K, V, SIZE, LOCK_FREE> {
+    pub fn cell_ref(&self) -> &'b Cell<K, V, LOCK_FREE> {
         self.cell_ref
     }
 
     /// Inserts a new key-value pair into the [`Cell`] without a uniqueness check.
     #[inline]
     pub fn insert(&'b self, key: K, value: V, partial_hash: u8, barrier: &'b Barrier) {
-        debug_assert!(!self.killed);
+        assert!(self.cell_ref.num_entries != u32::MAX, "array overflow");
 
-        if self.cell_ref.num_entries == u32::MAX {
-            panic!("array overflow");
-        }
-
-        let mut data_array_ptr = self.cell_ref.data.load(Relaxed, barrier);
+        let mut data_array_ptr = self.cell_ref.data_array.load(Relaxed, barrier);
         let data_array_head_ptr = data_array_ptr;
-        let preferred_index = partial_hash as usize % SIZE;
-        let mut free_index = SIZE;
+        let preferred_index = partial_hash as usize % ARRAY_SIZE;
+        let mut free_index = ARRAY_SIZE;
         while let Some(data_array_ref) = data_array_ptr.as_ref() {
             if (data_array_ref.occupied & (1_u32 << preferred_index)) == 0 {
                 free_index = preferred_index;
                 break;
             }
             let next_free_index = data_array_ref.occupied.trailing_ones() as usize;
-            if next_free_index < SIZE {
+            if next_free_index < ARRAY_SIZE {
                 free_index = next_free_index;
                 break;
             }
@@ -455,8 +408,8 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
             data_array_ptr = data_array_ref.link.load(Relaxed, barrier);
         }
 
-        if free_index == SIZE {
-            // Inserts a new DataArray at the head.
+        if free_index == ARRAY_SIZE {
+            // Inserts a new `DataArray` at the head.
             let mut new_data_array = Arc::new(DataArray::new());
             unsafe {
                 new_data_array.get_mut().unwrap().data[preferred_index]
@@ -473,16 +426,16 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
 
             new_data_array
                 .link
-                .swap((data_array_head_ptr.try_into_arc(), Tag::None), Relaxed);
+                .swap((data_array_head_ptr.get_arc(), Tag::None), Relaxed);
             self.cell_ref
-                .data
+                .data_array
                 .swap((Some(new_data_array), Tag::None), Release);
         } else {
             let data_array_ref = data_array_ptr.as_ref().unwrap();
             unsafe {
                 #[allow(clippy::cast_ref_to_mut)]
-                let data_array_mut_ref = &mut *(data_array_ref as *const DataArray<K, V, SIZE>
-                    as *mut DataArray<K, V, SIZE>);
+                let data_array_mut_ref =
+                    &mut *(data_array_ref as *const DataArray<K, V> as *mut DataArray<K, V>);
                 data_array_mut_ref.data[free_index]
                     .as_mut_ptr()
                     .write((key, value));
@@ -499,8 +452,8 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
     }
 
     /// Removes a key-value pair being pointed by the given [`EntryIterator`].
-    pub fn erase(&self, iterator: &mut EntryIterator<K, V, SIZE, LOCK_FREE>) -> Option<(K, V)> {
-        if self.killed || iterator.current_index == usize::MAX {
+    pub fn erase(&self, iterator: &mut EntryIterator<K, V, LOCK_FREE>) -> Option<(K, V)> {
+        if iterator.current_index == usize::MAX {
             return None;
         }
 
@@ -514,9 +467,8 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
             }
             self.num_entries_updated(self.cell_ref.num_entries - 1);
             #[allow(clippy::cast_ref_to_mut)]
-            let data_array_mut_ref = unsafe {
-                &mut *(data_array_ref as *const DataArray<K, V, SIZE> as *mut DataArray<K, V, SIZE>)
-            };
+            let data_array_mut_ref =
+                unsafe { &mut *(data_array_ref as *const DataArray<K, V> as *mut DataArray<K, V>) };
             let result = if LOCK_FREE {
                 data_array_mut_ref.removed |= 1_u32 << iterator.current_index;
                 None
@@ -538,7 +490,7 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
 
     /// Extracts the key-value pair being pointed by `self`.
     #[allow(clippy::unused_self)]
-    pub fn extract(&self, iterator: &mut EntryIterator<K, V, SIZE, LOCK_FREE>) -> (K, V) {
+    pub fn extract(&self, iterator: &mut EntryIterator<K, V, LOCK_FREE>) -> (K, V) {
         let data_array_ref = iterator.current_array_ptr.as_ref().unwrap();
         debug_assert!(!LOCK_FREE);
         debug_assert_ne!(
@@ -547,9 +499,8 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
         );
 
         #[allow(clippy::cast_ref_to_mut)]
-        let data_array_mut_ref = unsafe {
-            &mut *(data_array_ref as *const DataArray<K, V, SIZE> as *mut DataArray<K, V, SIZE>)
-        };
+        let data_array_mut_ref =
+            unsafe { &mut *(data_array_ref as *const DataArray<K, V> as *mut DataArray<K, V>) };
         data_array_mut_ref.occupied &= !(1_u32 << iterator.current_index);
         let entry_ptr = data_array_mut_ref.data[iterator.current_index].as_mut_ptr();
         #[allow(clippy::uninit_assumed_init)]
@@ -562,21 +513,21 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
 
     /// Tries to inherit the data array from the other [`Cell`].
     #[inline]
-    pub fn try_inherit(&self, other: &Locker<K, V, SIZE, LOCK_FREE>, barrier: &Barrier) -> bool {
-        if self.cell_ref.data.load(Relaxed, barrier).is_null() {
-            let other_data_array_ptr = other.cell_ref.data.load(Relaxed, barrier);
+    pub fn try_inherit(&self, other: &Locker<K, V, LOCK_FREE>, barrier: &Barrier) -> bool {
+        if self.cell_ref.data_array.load(Relaxed, barrier).is_null() {
+            let other_data_array_ptr = other.cell_ref.data_array.load(Relaxed, barrier);
             if let Some(other_data_array_ref) = other_data_array_ptr.as_ref() {
                 if other_data_array_ref.link.load(Relaxed, barrier).is_null() {
                     // The conditions are, `self` has none and `other` has a single data array.
-                    self.cell_ref.data.swap(
+                    self.cell_ref.data_array.swap(
                         (
-                            other.cell_ref.data.swap((None, Tag::None), Relaxed),
+                            other.cell_ref.data_array.swap((None, Tag::None), Relaxed),
                             Tag::None,
                         ),
                         Relaxed,
                     );
-                    debug_assert!(other.cell_ref.data.load(Relaxed, barrier).is_null());
-                    debug_assert!(!self.cell_ref.data.load(Relaxed, barrier).is_null());
+                    debug_assert!(other.cell_ref.data_array.load(Relaxed, barrier).is_null());
+                    debug_assert!(!self.cell_ref.data_array.load(Relaxed, barrier).is_null());
                     debug_assert_eq!(self.cell_ref.num_entries, 0);
                     self.num_entries_updated(1);
                     return true;
@@ -595,45 +546,39 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Locker<'b, K, V, SI
 
     /// Purges all the data.
     pub fn purge(&mut self, barrier: &Barrier) {
-        if !self.cell_ref.data.load(Relaxed, barrier).is_null() {
-            if let Some(data_array) = self.cell_ref.data.swap((None, Tag::None), Relaxed) {
+        self.cell_ref.state.fetch_or(KILLED, Release);
+        self.num_entries_updated(0);
+        if !self.cell_ref.data_array.load(Relaxed, barrier).is_null() {
+            if let Some(data_array) = self.cell_ref.data_array.swap((None, Tag::None), Relaxed) {
                 barrier.reclaim(data_array);
             }
         }
-        self.killed = true;
-        self.num_entries_updated(0);
     }
 
     /// Updates the number of entries.
     fn num_entries_updated(&self, num: u32) {
         #[allow(clippy::cast_ref_to_mut)]
         let cell_mut_ref =
-            unsafe { &mut *(self.cell_ref as *const _ as *mut Cell<K, V, SIZE, LOCK_FREE>) };
+            unsafe { &mut *(self.cell_ref as *const _ as *mut Cell<K, V, LOCK_FREE>) };
         cell_mut_ref.num_entries = num;
     }
 }
 
-impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> Drop
-    for Locker<'b, K, V, SIZE, LOCK_FREE>
-{
+impl<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Drop for Locker<'b, K, V, LOCK_FREE> {
     #[inline]
     fn drop(&mut self) {
         let mut current = self.cell_ref.state.load(Relaxed);
         loop {
             let wakeup = (current & WAITING) == WAITING;
-            let next = if self.killed {
-                KILLED | (current & (!(WAITING | LOCK)))
-            } else {
-                current & (!(WAITING | LOCK))
-            };
-            match self
-                .cell_ref
-                .state
-                .compare_exchange(current, next, Release, Relaxed)
-            {
+            match self.cell_ref.state.compare_exchange(
+                current,
+                current & (!(WAITING | LOCK)),
+                Release,
+                Relaxed,
+            ) {
                 Ok(_) => {
                     if wakeup {
-                        self.cell_ref.signal();
+                        self.cell_ref.wait_queue.signal();
                     }
                     break;
                 }
@@ -643,29 +588,27 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     }
 }
 
-pub struct Reader<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> {
-    cell_ref: &'b Cell<K, V, SIZE, LOCK_FREE>,
-    killed: bool,
+pub struct Reader<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> {
+    cell_ref: &'b Cell<K, V, LOCK_FREE>,
 }
 
-impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Reader<'b, K, V, SIZE, LOCK_FREE> {
+impl<'b, K: Eq, V, const LOCK_FREE: bool> Reader<'b, K, V, LOCK_FREE> {
     /// Locks the given [`Cell`].
     #[inline]
     pub fn lock(
-        cell: &'b Cell<K, V, SIZE, LOCK_FREE>,
+        cell: &'b Cell<K, V, LOCK_FREE>,
         barrier: &'b Barrier,
-    ) -> Option<Reader<'b, K, V, SIZE, LOCK_FREE>> {
+    ) -> Option<Reader<'b, K, V, LOCK_FREE>> {
         loop {
-            if let Some(reader) = Self::try_lock(cell, barrier)
-                .map_or_else(|| cell.wait(|| Self::try_lock(cell, barrier)), Some)
-            {
-                if reader.killed {
-                    return None;
-                }
-                return Some(reader);
+            if let Ok(reader) = Self::try_lock(cell, barrier) {
+                return reader;
             }
-            if cell.killed() {
-                return None;
+            if let Ok(reader) = cell.wait_queue.wait(|| {
+                // Marks that there is a waiting thread.
+                cell.state.fetch_or(WAITING, Release);
+                Self::try_lock(cell, barrier)
+            }) {
+                return reader;
             }
         }
     }
@@ -673,36 +616,35 @@ impl<'b, K: Eq, V, const SIZE: usize, const LOCK_FREE: bool> Reader<'b, K, V, SI
     /// Tries to lock the [`Cell`].
     #[inline]
     fn try_lock(
-        cell: &'b Cell<K, V, SIZE, LOCK_FREE>,
+        cell: &'b Cell<K, V, LOCK_FREE>,
         _barrier: &'b Barrier,
-    ) -> Option<Reader<'b, K, V, SIZE, LOCK_FREE>> {
+    ) -> Result<Option<Reader<'b, K, V, LOCK_FREE>>, ()> {
         let current = cell.state.load(Relaxed);
         if (current & LOCK_MASK) >= SLOCK_MAX {
-            return None;
+            return Err(());
+        }
+        if (current & KILLED) == KILLED {
+            return Ok(None);
         }
         if cell
             .state
             .compare_exchange(current, current + 1, Acquire, Relaxed)
             .is_ok()
         {
-            return Some(Reader {
-                cell_ref: cell,
-                killed: (current & KILLED) == KILLED,
-            });
+            Ok(Some(Reader { cell_ref: cell }))
+        } else {
+            Err(())
         }
-        None
     }
 
     /// Returns a reference to the [`Cell`].
     #[inline]
-    pub fn cell_ref(&self) -> &Cell<K, V, SIZE, LOCK_FREE> {
+    pub fn cell_ref(&self) -> &Cell<K, V, LOCK_FREE> {
         self.cell_ref
     }
 }
 
-impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool> Drop
-    for Reader<'b, K, V, SIZE, LOCK_FREE>
-{
+impl<'b, K: 'static + Eq, V: 'static, const LOCK_FREE: bool> Drop for Reader<'b, K, V, LOCK_FREE> {
     #[inline]
     fn drop(&mut self) {
         let mut current = self.cell_ref.state.load(Relaxed);
@@ -716,7 +658,7 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
             {
                 Ok(_) => {
                     if wakeup {
-                        self.cell_ref.signal();
+                        self.cell_ref.wait_queue.signal();
                     }
                     break;
                 }
@@ -726,31 +668,32 @@ impl<'b, K: 'static + Eq, V: 'static, const SIZE: usize, const LOCK_FREE: bool>
     }
 }
 
-pub struct DataArray<K: 'static + Eq, V: 'static, const SIZE: usize> {
-    link: AtomicArc<DataArray<K, V, SIZE>>,
+/// [`DataArray`] is a fixed size array of key-value pairs.
+pub struct DataArray<K: 'static + Eq, V: 'static> {
+    link: AtomicArc<DataArray<K, V>>,
     occupied: u32,
     removed: u32,
-    partial_hash_array: [u8; SIZE],
-    data: [MaybeUninit<(K, V)>; SIZE],
+    partial_hash_array: [u8; ARRAY_SIZE],
+    data: [MaybeUninit<(K, V)>; ARRAY_SIZE],
 }
 
-impl<K: 'static + Eq, V: 'static, const SIZE: usize> DataArray<K, V, SIZE> {
-    fn new() -> DataArray<K, V, SIZE> {
+impl<K: 'static + Eq, V: 'static> DataArray<K, V> {
+    fn new() -> DataArray<K, V> {
         DataArray {
             link: AtomicArc::null(),
             occupied: 0,
             removed: 0,
-            partial_hash_array: [0; SIZE],
+            partial_hash_array: [0_u8; ARRAY_SIZE],
             data: unsafe { MaybeUninit::uninit().assume_init() },
         }
     }
 }
 
-impl<K: 'static + Eq, V: 'static, const SIZE: usize> Drop for DataArray<K, V, SIZE> {
+impl<K: 'static + Eq, V: 'static> Drop for DataArray<K, V> {
     fn drop(&mut self) {
         let mut occupied = self.occupied;
         let mut index = occupied.trailing_zeros();
-        while (index as usize) < SIZE {
+        while (index as usize) < ARRAY_SIZE {
             let entry_mut_ptr = self.data[index as usize].as_mut_ptr();
             unsafe {
                 ptr::drop_in_place(entry_mut_ptr);
@@ -761,35 +704,6 @@ impl<K: 'static + Eq, V: 'static, const SIZE: usize> Drop for DataArray<K, V, SI
     }
 }
 
-struct WaitQueueEntry {
-    mutex: Mutex<u8>,
-    condvar: Condvar,
-    next_ptr: *mut WaitQueueEntry,
-}
-
-impl WaitQueueEntry {
-    fn new(next_ptr: *mut WaitQueueEntry) -> WaitQueueEntry {
-        WaitQueueEntry {
-            mutex: Mutex::new(0_u8),
-            condvar: Condvar::new(),
-            next_ptr,
-        }
-    }
-
-    fn wait(&self) {
-        let mut completed = self.mutex.lock().unwrap();
-        while *completed == 0_u8 {
-            completed = self.condvar.wait(completed).unwrap();
-        }
-    }
-
-    fn signal(&self) {
-        let mut completed = self.mutex.lock().unwrap();
-        *completed = 1_u8;
-        self.condvar.notify_one();
-    }
-}
-
 #[cfg(test)]
 mod test {
     use super::*;
@@ -799,10 +713,9 @@ mod test {
 
     #[test]
     fn cell_locker() {
-        const SIZE: usize = 32;
-        let num_threads = (SIZE * 2) as usize;
+        let num_threads = (ARRAY_SIZE * 2) as usize;
         let barrier = sync::Arc::new(sync::Barrier::new(num_threads));
-        let cell: sync::Arc<Cell<usize, usize, SIZE, true>> = sync::Arc::new(Cell::default());
+        let cell: sync::Arc<Cell<usize, usize, true>> = sync::Arc::new(Cell::default());
         let mut data: [u64; 128] = [0; 128];
         let mut thread_handles = Vec::with_capacity(num_threads);
         for thread_id in 0..num_threads {
@@ -826,7 +739,7 @@ mod test {
                         exclusive_locker.insert(
                             thread_id,
                             0,
-                            (thread_id % SIZE).try_into().unwrap(),
+                            (thread_id % ARRAY_SIZE).try_into().unwrap(),
                             &barrier,
                         );
                     } else {
@@ -835,7 +748,7 @@ mod test {
                                 .cell_ref()
                                 .search(
                                     &thread_id,
-                                    (thread_id % SIZE).try_into().unwrap(),
+                                    (thread_id % ARRAY_SIZE).try_into().unwrap(),
                                     &barrier
                                 )
                                 .unwrap(),
@@ -848,7 +761,11 @@ mod test {
                     assert_eq!(
                         read_locker
                             .cell_ref()
-                            .search(&thread_id, (thread_id % SIZE).try_into().unwrap(), &barrier)
+                            .search(
+                                &thread_id,
+                                (thread_id % ARRAY_SIZE).try_into().unwrap(),
+                                &barrier
+                            )
                             .unwrap(),
                         &(thread_id, 0_usize)
                     );
@@ -863,11 +780,11 @@ mod test {
         assert_eq!(cell.num_entries(), num_threads);
 
         let epoch_barrier = Barrier::new();
-        for thread_id in 0..SIZE {
+        for thread_id in 0..num_threads {
             assert_eq!(
                 cell.search(
                     &thread_id,
-                    (thread_id % SIZE).try_into().unwrap(),
+                    (thread_id % ARRAY_SIZE).try_into().unwrap(),
                     &epoch_barrier
                 ),
                 Some(&(thread_id, 0))
